# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
# NOTE: Uncomment if project is not installed (e.g., documentation is not
# generated with the help of tox.)
# sys.path.insert(0, os.path.abspath('../src'))


# -- Project information -----------------------------------------------------

project = 'ANTLeRinator'
author = 'Renata Hodovan, Akos Kiss'
copyright = '2017-2025, %s' % author

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
from antlerinator import __version__ as version
# The full version, including alpha/beta/rc tags.
release = version


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinxcontrib.runcmd',
    'inators.sphinx.argdoc',
]

# Add any paths that contain templates here, relative to this directory.
# templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.

html_theme = 'furo'

# The name for this set of Sphinx documents.
# "<project> v<release> documentation" by default.
#
html_title = project

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

# If true, links to the reST sources are added to the pages.
#
html_show_sourcelink = False


# -- Extension configuration -------------------------------------------------

intersphinx_mapping = {'python': ('https://docs.python.org/3', None)}



# Copyright (c) 2021-2025 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

import sys

from os import makedirs
from os.path import abspath, dirname, isfile, join

import pytest

from setuptools.dist import Distribution
from setuptools.errors import ModuleError

import antlerinator


is_windows = sys.platform.startswith('win32')
script_ext = '.bat' if is_windows else '.sh'

tested_antlr_version = '4.13.2'
resources_dir = join(dirname(abspath(__file__)), 'resources')

try:
    Distribution().get_command_class('editable_wheel')
    has_editable_wheel = True
except ModuleError:
    has_editable_wheel = False


def test_build_antlr_providers(tmpdir):
    """
    Test whether both the ``antlerinator:`` and the ``file:`` providers work.
    Also test whether ``build_antlr`` can deal with multiple ANTLR tool
    executions.
    """
    with tmpdir.as_cwd():
        antlr_jar_path = antlerinator.download(version=tested_antlr_version, path=join(str(tmpdir), 'antlr.jar'))

        dist = Distribution({
            'name': 'pkg',
            'script_name': 'setup.py',
            'script_args': ['build_antlr'],
            'options': {
                'build_antlr': {
                    'commands': f'''
                        antlerinator:{tested_antlr_version} {join(resources_dir, "Hello.g4")} -Dlanguage=Python3 -o {tmpdir} -Xexact-output-dir
                        file:{antlr_jar_path} {join(resources_dir, "Bello.g4")} -Dlanguage=Python3 -o {tmpdir} -Xexact-output-dir
                    ''',
                },
            },
        })

        dist.parse_command_line()
        dist.run_commands()

        assert isfile('HelloLexer.py')
        assert isfile('HelloParser.py')
        assert isfile('BelloLexer.py')
        assert isfile('BelloParser.py')


def test_build_antlr_java(tmpdir):
    """
    Test whether ``build_antlr`` can deal with a custom java VM.
    """
    with tmpdir.as_cwd():
        dist = Distribution({
            'name': 'pkg',
            'script_name': 'setup.py',
            'script_args': ['build_antlr'],
            'options': {
                'build_antlr': {
                    'commands': 'file:antlr.jar Dummy.g4',
                    'java': f'{join(resources_dir, "mock_java")}{script_ext}',
                },
            },
        })

        dist.parse_command_line()
        dist.run_commands()

        with open('mock_java_output.txt', 'r') as f:
            output = f.read()
        assert '-jar antlr.jar Dummy.g4' in output


def test_build(tmpdir):
    """
    Test whether lexer/parser generation happens when the general ``build``
    command is invoked (which is also invoked during ``install``).
    """
    with tmpdir.as_cwd():
        dist = Distribution({
            'name': 'pkg',
            'packages': ['pkg'],
            'script_name': 'setup.py',
            'script_args': ['build', f'--build-lib={join("build", "lib")}'],  # NOTE: --build-lib is necessary to ensure that purelib build directory is used
            'options': {
                'build_antlr': {
                    'commands': f'antlerinator:{tested_antlr_version} {join(resources_dir, "Hello.g4")} -Dlanguage=Python3 -o {join(str(tmpdir), "pkg")} -Xexact-output-dir',
                },
            },
        })
        makedirs('pkg')
        with open(join('pkg', '__init__.py'), 'w'):
            pass

        dist.parse_command_line()
        dist.run_commands()

        assert isfile(join('build', 'lib', 'pkg', 'HelloLexer.py'))
        assert isfile(join('build', 'lib', 'pkg', 'HelloParser.py'))


def test_develop(tmpdir):
    """
    Test whether lexer/parser generation happens for editable installs (via
    development mode).
    """
    with tmpdir.as_cwd():
        dist = Distribution({
            'name': 'pkg',
            'packages': ['pkg'],
            'script_name': 'setup.py',
            'script_args': ['develop'],
            'options': {
                'build_antlr': {
                    'commands': f'antlerinator:{tested_antlr_version} {join(resources_dir, "Hello.g4")} -Dlanguage=Python3 -o {join(str(tmpdir), "pkg")} -Xexact-output-dir',
                },
            },
        })
        makedirs('pkg')
        with open(join('pkg', '__init__.py'), 'w'):
            pass
        with open('setup.py', 'w') as f:
            f.write('from setuptools import setup; setup(name="pkg", packages=["pkg"])')

        dist.parse_command_line()
        dist.run_commands()

        assert isfile(join('pkg', 'HelloLexer.py'))
        assert isfile(join('pkg', 'HelloParser.py'))


@pytest.mark.skipif(not has_editable_wheel, reason='editable_wheel command unavailable')
def test_editable_wheel(tmpdir):
    """
    Test whether lexer/parser generation happens for editable installs (via
    editable wheels).
    """
    with tmpdir.as_cwd():
        dist = Distribution({
            'name': 'pkg',
            'packages': ['pkg'],
            'script_name': 'setup.py',
            'script_args': ['editable_wheel'],
            'options': {
                'build_antlr': {
                    'commands': f'antlerinator:{tested_antlr_version} {join(resources_dir, "Hello.g4")} -Dlanguage=Python3 -o {join(str(tmpdir), "pkg")} -Xexact-output-dir',
                },
            },
        })
        makedirs('pkg')
        with open(join('pkg', '__init__.py'), 'w'):
            pass

        dist.parse_command_line()
        dist.run_commands()

        assert isfile(join('pkg', 'HelloLexer.py'))
        assert isfile(join('pkg', 'HelloParser.py'))


def test_clean(tmpdir):
    """
    Test whether cleanup removes generated files.
    """
    with tmpdir.as_cwd():
        dist = Distribution({
            'name': 'pkg',
            'packages': ['pkg'],
            'script_name': 'setup.py',
            'script_args': ['clean'],
            'options': {
                'build_antlr': {
                    'output': join('pkg', 'Dummy*.py'),
                },
            },
        })
        makedirs('pkg')
        with open(join('pkg', '__init__.py'), 'w'):
            pass
        with open(join('pkg', 'DummyLexer.py'), 'w'):
            pass
        with open(join('pkg', 'DummyParser.py'), 'w'):
            pass

        dist.parse_command_line()
        dist.run_commands()

        assert isfile(join('pkg', '__init__.py'))
        assert not isfile(join('pkg', 'DummyLexer.py'))
        assert not isfile(join('pkg', 'DummyParser.py'))


def test_sdist(tmpdir):
    """
    Test whether generated files can be excluded from the sdist using
    MANIFEST.in.
    """
    with tmpdir.as_cwd():
        dist = Distribution({
            'name': 'pkg',
            'packages': ['pkg'],
            'script_name': 'setup.py',
            'options': {
                'build_antlr': {
                    'output': join('pkg', 'Dummy*.py'),
                },
            },
        })
        makedirs('pkg')
        with open(join('pkg', '__init__.py'), 'w'):
            pass
        with open(join('pkg', 'DummyLexer.py'), 'w'):
            pass
        with open(join('pkg', 'DummyParser.py'), 'w'):
            pass
        with open('MANIFEST.in', 'w') as f:
            f.write('exclude pkg/Dummy*.py')

        sdist = dist.get_command_class('sdist')
        cmd = sdist(dist)
        cmd.ensure_finalized()
        cmd.run()

        manifest = cmd.filelist.files
        assert join('pkg', '__init__.py') in manifest
        assert join('pkg', 'DummyLexer.py') not in manifest
        assert join('pkg', 'DummyParser.py') not in manifest



# Copyright (c) 2017-2023 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

import os
import subprocess
import sys

import pytest

import antlerinator


def run_antlr(jar_path):
    cmd = ('java', '-jar', jar_path)
    subprocess.run(cmd, check=True)


def run_download(args, exp_ok):
    cmd = (sys.executable, '-m', 'antlerinator') + args

    returncode = subprocess.run(cmd, check=False).returncode

    if exp_ok:
        assert returncode == 0
    else:
        assert returncode != 0


@pytest.mark.parametrize('antlr_version', [
    '4.6',  # the first ANTLR version that was supported by ANTLeRinator epoch 0
    # '4.7',
    # '4.7.1',
    # '4.7.2',
    # '4.8',
    # '4.9',
    # '4.9.1',
    '4.9.2',  # the last ANTLR version that was supported by ANTLeRinator epoch 0
])
@pytest.mark.parametrize('default_path', [
    True,
    False
])
class TestDownload:

    def test_cli(self, antlr_version, default_path, tmpdir):
        args = [f'--antlr-version={antlr_version}']

        if not default_path:
            jar_path = os.path.join(str(tmpdir), 'antlr4.jar')
            args += [f'--output={jar_path}']
        else:
            jar_path = antlerinator.default_antlr_jar_path(antlr_version)

        run_download(args=tuple(args) + ('--force', ), exp_ok=True)
        run_download(args=tuple(args) + ('--lazy', ), exp_ok=True)
        run_download(args=tuple(args), exp_ok=False)
        run_antlr(jar_path)

    def test_api(self, antlr_version, default_path, tmpdir):
        kwargs = {'version': antlr_version}

        if not default_path:
            kwargs['path'] = os.path.join(str(tmpdir), 'antlr4.jar')

        jar_path_force = antlerinator.download(force=True, **kwargs)
        jar_path_lazy = antlerinator.download(lazy=True, **kwargs)
        with pytest.raises(OSError):
            antlerinator.download(**kwargs)
        assert jar_path_force == jar_path_lazy
        run_antlr(jar_path_force)



# Copyright (c) 2017-2024 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

import contextlib
import errno
import ssl

from argparse import ArgumentParser
from importlib import metadata
from os import makedirs
from os.path import basename, dirname, exists, expanduser, join
from urllib.request import urlopen

import inators


__version__ = metadata.version(__package__)

try:
    __antlr_version__ = metadata.version('antlr4-python3-runtime')
except metadata.PackageNotFoundError:
    __antlr_version__ = None


def default_antlr_jar_path(version=None):
    """
    Default path to download the ANTLR v4 tool jar to.

    :param str version: The version of ANTLR v4. If ``None``, it defaults to the
        version of the installed antlr4 runtime package (unless the package is
        not installed, in which case a :exc:`ValueError` is raised).
    :return: The version-specific default path.
    """
    version = version or __antlr_version__
    if not version:
        raise ValueError('version must be specified if antlr4 runtime is not installed')
    return join(expanduser('~'), '.antlerinator', f'antlr-{version}-complete.jar')


def download(version=None, path=None, *, force=False, lazy=False):
    """
    Download the ANTLR v4 tool jar. (Raises :exc:`OSError` if jar is already
    available, unless ``lazy`` is ``True``.)

    :param str version: The version of ANTLR v4 tool jar to download. If
        ``None``, it defaults to the version of the installed antlr4 runtime
        package (unless the package is not installed, in which case a
        :exc:`ValueError` is raised).
    :param str path: Path to save the downloaded jar to. If ``None``, it
        defaults to ``default_antlr_jar_path(version)``.
    :param bool force: Force download even if jar already exists at path.
    :param bool lazy: Don't report an error if jar already exists at path and
        don't try to download it either.
    :return: Path to the downloaded jar.
    """

    default_tool_path = default_antlr_jar_path(version)
    tool_path = path or default_tool_path
    tool_url = f'https://www.antlr.org/download/{basename(default_tool_path)}'

    if exists(tool_path):
        if lazy:
            return tool_path
        if not force:
            raise OSError(errno.EEXIST, 'file already exists', tool_path)

    ssl_context = ssl.create_default_context(purpose=ssl.Purpose.SERVER_AUTH)
    with contextlib.closing(urlopen(tool_url, context=ssl_context)) as response:
        tool_bytes = response.read()

    tool_dir = dirname(tool_path)
    makedirs(tool_dir, exist_ok=True)

    with open(tool_path, mode='wb') as tool_file:
        tool_file.write(tool_bytes)

    return tool_path


def execute():
    """
    Entry point of the download helper tool that eases getting the right
    version of the ANTLR v4 tool jar.
    """

    arg_parser = ArgumentParser(description='Download helper tool to get the right version of the ANTLR v4 tool jar.')

    arg_parser.add_argument('--antlr-version', metavar='VERSION', default=__antlr_version__,
                            help='version of ANTLR v4 tool jar to download (default: %(default)s)')
    arg_parser.add_argument('--output', metavar='FILE', default=None,
                            help=f'path to save the downloaded jar to (default: {default_antlr_jar_path("VERSION").replace(expanduser("~"), "~")})')

    mode_group = arg_parser.add_mutually_exclusive_group()
    mode_group.add_argument('--force', action='store_true', default=False,
                            help='force download even if jar already exists at the output path')
    mode_group.add_argument('--lazy', action='store_true', default=False,
                            help='don\'t report an error if jar already exists at the output path and don\'t try to download it either')

    inators.arg.add_version_argument(arg_parser, version=__version__)

    args = arg_parser.parse_args()

    download(version=args.antlr_version, path=args.output, force=args.force, lazy=args.lazy)



# Copyright (c) 2021-2024 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

import glob
import os
import re
import shlex
import sys

from setuptools import Command
from setuptools.errors import ModuleError, OptionError

from .download import download


class build_antlr(Command):

    description = 'build (generate) parsers/lexers from ANTLR v4 grammar files'

    user_options = [
        ('commands=', None, 'list of antlr4 invocations'),
        ('output=', None, 'list of file names or glob patterns that are the result of commands'),
        ('java=', None, 'path to java'),
    ]

    def initialize_options(self):
        self.commands = None
        self.output = None
        self.java = None

    def finalize_options(self):
        # parse 'commands' option
        commands = self.commands or []

        if isinstance(commands, str):
            if '\n' in commands:
                commands = commands.splitlines()
            else:
                commands = commands.split(',')
            commands = [cmd.strip() for cmd in commands if cmd.strip()]

        if not isinstance(commands, list):
            raise OptionError(f"'commands' must be a list of strings or tuples (got {commands!r})")

        cmd_re = re.compile('^(?P<provider>^[^\\d\\W]\\w*):(?P<provider_arg>\\S*)\\s+(?P<antlr_args>.*)$')
        providers = {
            'file': lambda arg: arg,
            'antlerinator': lambda arg: download(arg, lazy=True),
        }
        posix = sys.platform != 'win32'

        for i, cmd in enumerate(commands):
            if isinstance(cmd, str):
                m = cmd_re.match(cmd)
                if not m:
                    raise OptionError(f"strings in 'commands' must start with a 'provider:arg' pattern (got {cmd!r})")
                provider, provider_arg, antlr_args = m.group('provider', 'provider_arg', 'antlr_args')

                provider = providers.get(provider)
                if not provider:
                    raise OptionError(f"unknown provider in 'commands' (options: {', '.join(providers.keys())}; got: {provider})")
                antlr_args = tuple(shlex.split(antlr_args, posix=posix))

                cmd = (provider, provider_arg, antlr_args)

            if isinstance(cmd, tuple):
                if len(cmd) != 3:
                    raise OptionError(f"tuples in 'commands' must be 3-tuples (got {cmd!r})")
            else:
                raise OptionError(f"elements in 'commands' must be strings or tuples (got {commands!r})")

            commands[i] = cmd

        self.commands = commands

        # process 'output' option
        self.ensure_string_list('output')
        if self.output is None:
            self.output = []

        # ensure default for 'java' option
        if not self.java:
            self.java = 'java'

    def run(self):
        for provider, provider_arg, antlr_args in self.commands:
            self.spawn([self.java, '-jar', provider(provider_arg)] + list(antlr_args))


class clean_antlr(Command):

    description = 'clean parsers/lexers generated from ANTLR v4 grammar files'

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        output = self.get_finalized_command('build_antlr').output
        files = [f for o in output for f in glob.glob(o, recursive=True)]
        for f in files:
            self.execute(os.unlink, (f,))


def register(dist):

    build = dist.get_command_class('build')
    develop = dist.get_command_class('develop')
    clean = dist.get_command_class('clean')

    class antlerinator_build(build):
        def run(self):
            self.run_command('build_antlr')
            build.run(self)

    class antlerinator_develop(develop):
        def run(self):
            self.run_command('build_antlr')
            develop.run(self)

    class antlerinator_clean(clean):
        def run(self):
            self.run_command('clean_antlr')
            clean.run(self)

    dist.cmdclass['build'] = antlerinator_build
    dist.cmdclass['develop'] = antlerinator_develop
    dist.cmdclass['clean'] = antlerinator_clean

    # Patch editable_wheel only if available
    try:
        editable_wheel = dist.get_command_class('editable_wheel')

        class antlerinator_editable_wheel(editable_wheel):
            def run(self):
                self.run_command('build_antlr')
                editable_wheel.run(self)

        dist.cmdclass['editable_wheel'] = antlerinator_editable_wheel
    except ModuleError:
        pass



# Copyright (c) 2021-2022 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

from inators.arg import add_argument

from .download import __antlr_version__, default_antlr_jar_path, download


def add_antlr_argument(
        parser,
        short_alias=(),
        long_alias=(),
        *,
        metavar='FILE',
        help=f'path of the ANTLR v4 tool jar file (default: {default_antlr_jar_path(__antlr_version__ or "VERSION")})'
):
    """
    add_antlr_argument(parser, short_alias=(), long_alias=(), *, metavar='FILE', help='path of the ANTLR v4 tool jar file (default:  ~/.antlerinator/antlr-VERSION-complete.jar)')

    Add an ``--antlr`` command-line argument to ``parser``.

    The default processing of the added argument is implemented in
    :func:`process_antlr_argument`.

    :param ~argparse.ArgumentParser parser: The parser to add the argument to.
    :param short_alias: Add short flag alias(es) for ``--antlr``.
    :type short_alias: str or list(str) or tuple(str)
    :param long_alias: Add long option alias(es) for ``--antlr``.
    :type long_alias: str or list(str) or tuple(str)
    :param str metavar: Override the default argument name in usage messages.
    :param str help: Override the default description of ``--antlr``.

    .. argdoc {'help': 'path of the ANTLR v4 tool jar file (default:  ~/.antlerinator/antlr-VERSION-complete.jar)'}
    """
    # NOTE: The first line of the docstring (i.e., the documented signature)
    #   must be kept in sync with the actual signature of the function!

    add_argument(parser, short_alias, '--antlr', long_alias,
                 metavar=metavar, default=None, help=help)


def process_antlr_argument(args):
    """
    Lazily download the ANTLR v4 tool jar to the default path if ``--antlr`` was
    *not* given on the command line. I.e., download and copy the jar to the
    default path if it is not already there (using
    :func:`antlerinator.download`). Also set ``args.antlr`` to the path to the
    jar. No-op if ``--antlr`` was specified on the command line.

    This implements the default processing of the ``--antlr`` command-line
    argument added by :func:`add_antlr_argument`.

    :param args: A namespace object populated by
        :meth:`argparse.ArgumentParser.parse_args`.
    """

    if not args.antlr:
        args.antlr = download(lazy=True)



# Copyright (c) 2021 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

import argparse
import pytest

import antlerinator


@pytest.mark.parametrize('func_args, func_kwargs, sys_argv, exp', [
    ([], {}, [], None),
    ([], {}, ['--antlr', './antlr.jar'], './antlr.jar'),
    (['-A'], {}, ['-A', './antlr.jar'], './antlr.jar'),
    ([(), '--antlr:antlr'], {}, ['--antlr:antlr', './antlr.jar'], './antlr.jar'),
    ([], {'short_alias': '-A'}, ['-A', './antlr.jar'], './antlr.jar'),
    ([], {'long_alias': '--antlr:antlr'}, ['--antlr:antlr', './antlr.jar'], './antlr.jar'),
])
def test_add_antlr_argument(func_args, func_kwargs, sys_argv, exp):
    parser = argparse.ArgumentParser()
    antlerinator.add_antlr_argument(parser, *func_args, **func_kwargs)
    args = parser.parse_args(sys_argv)
    assert args.antlr == exp
